#include <assert.h>
#include <math.h>
#include "unit.h"
#include "common.h"
#include "game.h"
#include "map.h"

#define TEST_ADD_BLOCKED_MAX_COUNT 10
#define TEST_BLOCKED_WAIT_MAX (FPS*5)

struct data_test
{
	int last_draw_x, last_draw_y;
	/*
		These two depend on the value of unit.is_going:
		1: going_to: the target position
		   path: NULL or the path the unit is using
		0: going_to: <anything>
			path: NULL
	*/
	struct tpos going_to;
	struct tpos *path;
	unsigned int path_i; //the last used index in path[]
	/*
		0: not waiting
		1-255: count of waited frames
	*/
	uint8_t blocked_wait_counter;
	/*
		"additional blocked positions", for pathfinding.
		You should quite easily be able to control the cleverness of the unit
		by varying the maximum size of this.
	*/
	struct tpos_linked *additional_blocked;
	/*
		1: The value for get_target_reached is in going_to.
		0: Target not reached.
	*/
	uint8_t target_reached;
	
	uint8_t make_them_evade;

	uint8_t count_of_units_going_to_same_place;
};

int load_test(struct unit *unit);
void unload_test(struct unit *unit);
int get_draw_info_test(struct unit *unit, struct draw_image_info *images, unsigned int maxcount);
int get_last_draw_rects_test(struct unit *unit, SDL_Rect *rects, unsigned int maxcount);
int set_tpos_test(struct unit *unit, int x, int y, int slot);
int get_blocked_tiles_test(struct unit *unit, struct tpos *ts);
int action_test3(struct unit *unit, struct game *game);
void get_select_area_test(struct unit *unit, SDL_Rect *rect);
int click_map_test(struct unit *unit, int tx, int ty, unsigned int selection_size);
int click_unit_test(struct unit *unit, int ui);
int going_to_test(struct unit *unit, struct tpos *t);
struct tpos * get_copy_of_path_test(struct unit *unit, struct tpos *starting_from);
struct tpos_linked * get_additional_blocked_test(struct unit *unit);
void sync_additional_blocked_test(struct unit *unit, struct tpos_linked *a);
int get_target_reached_test(struct unit *unit, struct tpos *t);
int evade_test(struct unit *unit, struct game *game, struct unit *caller, unsigned int recursion_depth);

//REMEMBER TO ADD THE FUNCTIONS YOU ADD ABOVE TO THE ARRAY BELOW

int8_t blockedfunc_test(struct tile_info *info);

struct unit_spec unit_specs[] =
{
	{
		"test_fast", {"test3_0", "test3_1", "test3_2", "test3_3", "test3_4", "test3_5", "test3_6", "test3_7"},
		&load_test,
		&unload_test,
		&get_draw_info_test,
		&get_last_draw_rects_test,
		&set_tpos_test,
		&get_blocked_tiles_test,
		&action_test3,
		&get_select_area_test,
		&click_map_test,
		&click_unit_test,
		&going_to_test,
		&get_copy_of_path_test,
		&get_additional_blocked_test,
		&sync_additional_blocked_test,
		&get_target_reached_test,
		&evade_test,
	},
	{
		"test_slow", {"test2"},
		&load_test,
		&unload_test,
		&get_draw_info_test,
		&get_last_draw_rects_test,
		&set_tpos_test,
		&get_blocked_tiles_test,
		&action_test3,
		&get_select_area_test,
		&click_map_test,
		&click_unit_test,
		&going_to_test,
		&get_copy_of_path_test,
		&get_additional_blocked_test,
		&sync_additional_blocked_test,
		&get_target_reached_test,
		&evade_test,
	},
};

void unit_specs_initialize(struct imglib *imglib)
{
	struct unit_spec *p = unit_specs;
	unsigned int i;
	for(i=0;i<sizeof(unit_specs)/sizeof(struct unit_spec); i++)
	{
		fprintf(stderr, "unit_specs_initialize initializing unit_spec \"%s\"\n", p->name);
		unsigned int i;
		for(i=0; i<UNIT_IMAGE_MAX_COUNT; i++)
		{
			if(p->imagenames[i] == NULL){
				p->images[i] = NULL;
				continue;
			}
			p->images[i] = imglib_get_image(imglib, p->imagenames[i]);
			if(p->images[i] == NULL){
				fprintf(stderr, "unit_specs_initialize: couldn't imglib_get_image \"%s\"\n", p->imagenames[i]);
			}
			else{
				fprintf(stderr, "unit_specs_initialize: loaded image %i=\"%s\" -> pointer %i\n", i, p->imagenames[i], (int)p->images[i]);
			}
		}
		p++;
	}
}

//=== unit specs ===

int load_test(struct unit *unit)
{
	//fprintf(stderr, "load_test()\n");
	unit->data = (struct data_test *)malloc(sizeof(struct data_test));
	if(unit->data == NULL){
		fprintf(stderr, "load_test(): couldn't allocate data\n");
		return -1;
	}
	struct data_test *data = (struct data_test *)unit->data;
	data->last_draw_x = 0;
	data->last_draw_y = 0;
	data->path = NULL;
	data->additional_blocked = NULL;
	data->target_reached = 0;
	data->make_them_evade = 0;
	unit->direction = DIRECTION_R;
	if(strcmp(unit_specs[unit->type_id].name, "test_slow")){
		unit->move_speed = 4;
	}
	else if(strcmp(unit_specs[unit->type_id].name, "test_fast")){
		unit->move_speed = 0.5;
	}
	else assert(0);
	return 0;
}
void unload_test(struct unit *unit)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	if(data->path) free(data->path);
	tpos_linked_list_clear(&data->additional_blocked);
	free(unit->data);
}
int get_draw_info_test(struct unit *unit, struct draw_image_info *images, unsigned int maxcount)
{
	if(maxcount < 1) return 0;
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	if(unit->direction >= 0 && unit->direction <= 7){
		images[0].s = unit_specs[unit->type_id].images[0];
		//the direction value isn't meant to be used for this
		//images[0].s = unit_specs[unit->type_id].images[unit->direction];
	}
	else{
		images[0].s = unit_specs[unit->type_id].images[0];
	}
	images[0].px = unit->p.x;
	images[0].py = unit->p.y;
	data->last_draw_x = unit->p.x;
	data->last_draw_y = unit->p.y;
	return 1;
}
int get_last_draw_rects_test(struct unit *unit, SDL_Rect *rects, unsigned int maxcount)
{
	if(maxcount < 1) return 0;
	assert(unit->data!=NULL);
	if(unit_specs[unit->type_id].images[0] == NULL) return 0;
	struct data_test *data = (struct data_test *)unit->data;
	//fprintf(stderr, "get_last_draw_rects_test: x=%i, y=%i\n", data->last_draw_x, data->last_draw_y);
	rects[0].x = data->last_draw_x;
	rects[0].y = data->last_draw_y;
	rects[0].w = unit_specs[unit->type_id].images[0]->w;
	rects[0].h = unit_specs[unit->type_id].images[0]->h;
	return 1;
}
int set_tpos_test(struct unit *unit, int x, int y, int slot)
{
	//fprintf(stderr, "set_tpos_test(%i, %i, %i, %i)\n", (int)unit, x, y, slot);
	unit->p.x = x * MAP_TILE_W;
	unit->p.y = y * MAP_TILE_H;
	unit->graphics_changed = 1;
	return 0;
}
int get_blocked_tiles_test(struct unit *unit, struct tpos *ts)
{
	//assert(unit->data!=NULL);
	//struct data_test *data = (struct data_test *)unit->data;
	if(unit->is_moving){
		ts[0].x = unit->moving_to.x / MAP_TILE_W;
		ts[0].y = unit->moving_to.y / MAP_TILE_H;
		ts[1].x = ts[0].x;
		ts[1].y = ts[0].y;
		tadd_direction(&ts[1], direction_reverse(unit->direction));
		return 2;
	}
	else{
		ts[0].x = unit->p.x / MAP_TILE_W;
		ts[0].y = unit->p.y / MAP_TILE_H;
		return 1;
	}
}
#define sel_fprintf(...) {if(unit->selected)fprintf(__VA_ARGS__);}
int action_test3(struct unit *unit, struct game *game)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	
	struct tpos ct; //current tpos
	ct.x = unit->p.x/MAP_TILE_W;
	ct.y = unit->p.y/MAP_TILE_H;

action_test3_restart:
	//if unit is moving at the moment, there's nothing to do here
	if(unit->is_moving) return 0;

	//otherwise it should try to get moving
	
	if(!unit->is_going) return 0;

	//sel_fprintf(stderr, "action_test3: UNIT %i AT (%i,%i) TO (%i,%i)\n", unit->index, ct.x, ct.y, data->going_to.x, data->going_to.y);
	sel_fprintf(stderr, "UNIT %i AT (%i,%i) TO (%i,%i)\n", unit->index, ct.x, ct.y, data->going_to.x, data->going_to.y);

	//if the goal is reached, stop and free the path
	if(ct.x == data->going_to.x && ct.y == data->going_to.y)
	{
		sel_fprintf(stderr, "Target position reached.\n");
		tpos_linked_list_clear(&data->additional_blocked);
		if(data->path) free(data->path);
		data->path = NULL;
		unit->is_going = 0;
		//unit->direction = DIRECTION_NONE;
		data->target_reached = 1;
		return 0;
	}

	//if the unit is going to somwhere but hasn't got a path, try to make one
	if(unit->is_going && data->path==NULL){
		data->blocked_wait_counter = 0;
		sel_fprintf(stderr, "Has no path -> finding path... ");
		//a path needs to be generated
		data->path = game_find_path(game, ct.x, ct.y, data->going_to.x, data->going_to.y, &blockedfunc_test, data->additional_blocked);
		if(data->path != NULL){
			data->path_i = 0;
			sel_fprintf(stderr, "path found.\n");
		}
		else{
			sel_fprintf(stderr, "path not found ");
			if(data->make_them_evade == 0){
				sel_fprintf(stderr, "-> clearing additional_blocked, enabling make_them_evade\n");
				tpos_linked_list_clear(&data->additional_blocked),
				data->make_them_evade = 1;
				goto action_test3_restart;
			}
			sel_fprintf(stderr, "-> giving up.\n");
			//TODO: keep trying?
			unit->is_going = 0;
			//unit->direction = DIRECTION_NONE;
		}
	}

	//If unit is not going anywhere or has no path, there's nothing to do.
	if(unit->is_going == 0 || data->path == NULL) return 0;

	//see what is ahead
	int path_i_next = data->path_i+1;
	struct tpos nt;
	nt.x = data->path[path_i_next].x;
	nt.y = data->path[path_i_next].y;
	sel_fprintf(stderr, "Next tpos in planned path: (%i,%i): ", nt.x, nt.y);
	struct tile_info ti;
	if(game_get_tile_info(game, &ti, nt.x, nt.y) == -1){
		//failed, so cancel the whole thing
		fprintf(stderr, "action_test3: ERROR: can't get info for next path position\n");
		tpos_linked_list_clear(&data->additional_blocked);
		free(data->path);
		data->path = NULL;
		unit->is_going = 0;
		//unit->direction = DIRECTION_NONE;
		assert(0); //DEBUG
		return -1;
	}

	//If the path is diagonal, check if the sides of the path are blocked

	/*unsigned int sides_are_clear = 1;
	
	//path is diagonal <=> new x != current x and new y != current y
	if(nt.x != ct.x && nt.y != ct.y){
		sel_fprintf(stderr, "path is diagonal...\n");
		int dx = nt.x - ct.x;
		int dy = nt.y - ct.y;
		assert(abs(dx)==1); //path is invalid if not
		assert(abs(dy)==1); //path is invalid if not
		struct tpos t1, t2;
		t1.x = ct.x + dx;
		t1.y = ct.y;
		t2.x = ct.x;
		t2.y = ct.y + dy;
		struct tile_info i1, i2;
		int ii1 = game_get_tile_info(game, &i1, t1.x, t1.y);
		int ii2 = game_get_tile_info(game, &i2, t2.x, t2.y);
		if(ii1 == -1 || ii2 == -1){
			//failed, so cancel the whole thing
			fprintf(stderr, "action_test3: ERROR: can't get info for path's side position\n");
			tpos_linked_list_clear(&data->additional_blocked);
			free(data->path);
			data->path = NULL;
			unit->is_going = 0;
			//unit->direction = DIRECTION_NONE;
			assert(0); //DEBUG
			return -1;
		}
		if(i1.unit_count != 0 || i2.unit_count != 0){
			sides_are_clear = 0;
		}
		unsigned int i;
		unsigned int staying1=0;
		for(i=0; i<i1.unit_count; i++){
			struct unit *u = &game->units[i1.unit_i[i]];
			assert(i1.unit_i[i] != unit->index);
			if(u->is_going == 0){
				if(data->make_them_evade){
					if(unit_evade(u, game, unit, 0)==-1){
						staying1 = 1;
						break;
					}
				}
				else{
					staying1 = 1;
					break;
				}
			}
		}
		unsigned int staying2=0;
		for(i=0; i<i2.unit_count; i++){
			struct unit *u = &game->units[i2.unit_i[i]];
			assert(i2.unit_i[i] != unit->index);
			if(u->is_going == 0){
				if(data->make_them_evade){
					if(unit_evade(u, game, unit, 0)==-1){
						staying2 = 1;
						break;
					}
				}
				else{
					staying2 = 1;
					break;
				}
			}
		}
		if(staying1 && data->make_them_evade==0){
			unsigned int count=tpos_linked_list_count(data->additional_blocked);
			if(count < TEST_ADD_BLOCKED_MAX_COUNT){
				//add it to the beginning of additional_blocked
				tpos_linked_list_add_single_copy(&data->additional_blocked,&t1);
				//get a new path:
				if(data->path) free(data->path);
				data->path = NULL;
			}
		}
		if(staying2 && data->make_them_evade==0){
			unsigned int count=tpos_linked_list_count(data->additional_blocked);
			if(count < TEST_ADD_BLOCKED_MAX_COUNT){
				//add it to the beginning of additional_blocked
				tpos_linked_list_add_single_copy(&data->additional_blocked,&t2);
				//get a new path:
				if(data->path) free(data->path);
				data->path = NULL;
			}
		}
		if(staying1 || staying2){
			if(data->path==NULL) goto action_test3_restart;
			//if not, nothing else to do
			unit->is_going = 0;
			tpos_linked_list_clear(&data->additional_blocked);
			free(data->path);
			data->path = NULL;
			//unit->direction = DIRECTION_NONE;
			return 0;
		}
	}
	*/
	/*
		If there are no units in where this unit is going.
		If the path is diagonal, there are no units at the sides, start
		moving there.
	*/

	if(ti.unit_count == 0 /*&& sides_are_clear*/){
		sel_fprintf(stderr, "no units.\n");
		unit->moving_to.x = nt.x * MAP_TILE_W;
		unit->moving_to.y = nt.y * MAP_TILE_H;
		unit->is_moving = 1;
		unit->direction = tjump_direction(&data->path[data->path_i], &data->path[path_i_next]);

		sel_fprintf(stderr, "There are no other units in here -> moving to (%i,%i)t = (%i,%i)px\n", nt.x, nt.y, unit->moving_to.x, unit->moving_to.y);
		sel_fprintf(stderr, "direction set to %s.\n", direction_string(unit->direction));
		
		data->path_i = path_i_next;
		//sel_fprintf(stderr, "unit->direction=%i\n", unit->direction);
		data->blocked_wait_counter = 0;
		return 0;
	}

	unsigned int all_going_away = 1;
	unsigned int all_going_to_same_place = 1;
	unsigned int some_going_to_same_place = 0;

	/*
		There are units at where we are going.
		Now set the above values accordingly.
	*/

	//sel_fprintf(stderr, "there are units in the next position (%i,%i):\n", nt.x, nt.y);
	sel_fprintf(stderr, "units:\n");
	
	unsigned int i;
	for(i=0; i<ti.unit_count; i++){
		struct unit *u = &game->units[ti.unit_i[i]];
		sel_fprintf(stderr, "    unit %i: %s: ", ti.unit_i[i], direction_string(u->direction));
		if(ti.unit_i[i] == unit->index){
			fprintf(stderr, "action_test3: ERROR: The current unit is at target position.\n");
			assert(0);
			return -1;
		}
		
		if(u->is_going == 0){
			/*
				Check if it has reached our target position.
				If it has, then we're maybe reached it too.
			*/
			unsigned int has_reached_our_target = 0;
			struct tpos ta;
			if(unit_get_target_reached(u, &ta) == 0){
				if(ta.x == data->going_to.x && ta.y == data->going_to.y){
					has_reached_our_target = 1;
				}
			}
			if(!has_reached_our_target){
				sel_fprintf(stderr, "is staying and has had different or no target\n");
				if(data->make_them_evade){
					if(unit_evade(u, game, unit, 0)==0){
						continue;
					}
				}
				all_going_away = 0;
				all_going_to_same_place = 0;
				continue;
			}
			else{
				sel_fprintf(stderr, "has reached our target position.\n");
				unsigned int count = data->count_of_units_going_to_same_place;
				//unsigned int count = game_get_number_of_units_going_to(game, data->going_to.x, data->going_to.y);
				float d = distance(ct.x, ct.y, data->going_to.x, data->going_to.y);
				float e;
				if(game->settings.reach_target_always_when_other_reached_it_too){
					e = 1000;
				}
				else{
					e = (sqrtf((float)count)+1)/1.25;
				}
				sel_fprintf(stderr, "distance=%f, count=%i, e=%f\n", d, count, e);
				if(d < e){
					sel_fprintf(stderr, "Distance short enough -> that's \"target position reached\" for us, too.\n");
					unit->is_going = 0;
					//unit->direction = DIRECTION_NONE;
					data->target_reached = 1;
					tpos_linked_list_clear(&data->additional_blocked);
					if(data->path) free(data->path);
					data->path = NULL;
					return 0;
				}
				sel_fprintf(stderr, "But the distance from here is too long.\n");
				all_going_away = 0;
				all_going_to_same_place = 0;
			}
		}
		else{
			if(u->direction != DIRECTION_NONE){
				struct tpos s;
				s.x = nt.x;
				s.y = nt.y;
				tadd_direction(&s, u->direction);
				
				if(s.x == ct.x && s.y == ct.y){
					sel_fprintf(stderr, "going to here: ");
					all_going_away = 0;
				}
				else{
					sel_fprintf(stderr, "going to (%i,%i): ",s.x, s.y);
				}
			}
			else{
				sel_fprintf(stderr, "is_going, but doesn't know direction: ");
				all_going_away = 0;
			}
		}

		//Check where the other unit is going

		struct tpos ut;
		if(unit_going_to(u, &ut) == -1){
			//The other unit doesn't tell where it's going
			sel_fprintf(stderr, "doesn't tell where it's going -> all_going_to_same_place = 0\n");
			all_going_to_same_place = 0;
			continue;
		}
		
		//sel_fprintf(stderr, "it's going to (%i,%i): ", ut.x, ut.y);

		//Check if the other unit isn't going to the same place as this

		if(ut.x != data->going_to.x || ut.y != data->going_to.y){
			sel_fprintf(stderr, "is not going to the same place -> all_going_to_same_place = 0\n");
			//unit is going somewhere else
			all_going_to_same_place = 0;
			continue;
		}

		//Chances are the unit is going to the same place as this

		some_going_to_same_place = 1;
		
		sel_fprintf(stderr, "it's going to the same place as this unit.\n");
		sel_fprintf(stderr, "        Synchronizing additional_blocked to this unit.\n");

		//Synchronize unit's additional_blocked to this unit
		unit_sync_additional_blocked(unit, unit_get_additional_blocked(u));
	}

	/*
		TODO:
		If there are other units at the target position going to the same
		place as this,
		1) Synchronize our additional_blocked to all of them.
		2) Clear the path of every unit which had something added to their
		   additional_blocked (including this)
		After those, we're done. TODO: return 0;
	*/

	if(some_going_to_same_place){
		unsigned int i;
		for(i=0; i<ti.unit_count; i++){
			struct unit *u = &game->units[ti.unit_i[i]];
			struct tpos ut;
			//discard units that are not going to the same place as this
			if(unit_going_to(u, &ut) == -1){
				continue;
			}
			sel_fprintf(stderr, "Synchronizing additional_blocked of this unit to unit %i.\n", u->index);
			unit_sync_additional_blocked(u, data->additional_blocked);
			if(data->path==NULL) goto action_test3_restart;
		}
		//if(data->path==NULL) goto action_test3_restart;
		if(all_going_to_same_place) return 0;
	}
	
	/*
		The if above should have got all of these.
		(unit_sync_additional_blocked can clear the path)
	*/
	assert(data->path != NULL);
	
	//If all units in the target position are going away, wait for them to go.

	if(all_going_away){
		sel_fprintf(stderr, "None are staying -> setting direction towards them and waiting\n");
		unit->direction = tjump_direction(&data->path[data->path_i], &data->path[path_i_next]);

		//Wait more if not waited the maximum time already.
		if(data->blocked_wait_counter < TEST_BLOCKED_WAIT_MAX){
			data->blocked_wait_counter++;
			return 0;
		}
		sel_fprintf(stderr, "Waited enough, nt.ying something else\n");
		data->blocked_wait_counter = 0;
		/*
			If the unit has been waiting long enough, continue.
			The position gets marked blocked and a new path is generated.
		*/
	}
	
	if(all_going_to_same_place){
		sel_fprintf(stderr, "All are going to same place as this unit\n");
		//TODO
		//sel_fprintf(stderr, "We'll just wait.\n");
		sel_fprintf(stderr, "Clearing the path.\n");
		free(data->path);
		data->path = NULL;
		return 0;
	}

	/*
		There's nothing clever we can do any more, so just set the position
		blocked and attempt to make a new path.
	*/

	unsigned int count = tpos_linked_list_count(data->additional_blocked);
	
	if(count < TEST_ADD_BLOCKED_MAX_COUNT){
		sel_fprintf(stderr, "Clearing the path and adding (%i,%i) to additional_blocked\n", nt.x, nt.y);

		//add it to the beginning of additional_blocked
		tpos_linked_list_add_single_copy(&data->additional_blocked, &nt);
		
		//get a new path:
		free(data->path);
		data->path = NULL;
		goto action_test3_restart;
	}
	
	sel_fprintf(stderr, "Some are staying and additional_blocked is full. Trying to make them evade.\n");
	
	unsigned int all_evade = 1;
	for(i=0; i<ti.unit_count; i++){
		struct unit *u = &game->units[ti.unit_i[i]];
		if(unit_evade(u, game, unit, 0) == -1){
			all_evade = 0;
			break;
		}
	}

	if(all_evade){
		sel_fprintf(stderr, "Could get all units at target position to evade -> now waiting.\n");
		return 0;
	}

	//Well, it can get worse, if the blocked list is alredy full. Now give up.
	sel_fprintf(stderr, "Couldn't get all units at target position to evade -> giving up.\n");
	unit->is_going = 0;
	tpos_linked_list_clear(&data->additional_blocked);
	free(data->path);
	data->path = NULL;
	//unit->direction = DIRECTION_NONE;

	return 0;
}
void get_select_area_test(struct unit *unit, SDL_Rect *rect)
{
	//assert(unit->data!=NULL);
	//struct data_test *data = (struct data_test *)unit->data;
	rect->x = unit->p.x;
	rect->y = unit->p.y;
	rect->w = MAP_TILE_W;
	rect->h = MAP_TILE_H;
}
int click_map_test(struct unit *unit, int tx, int ty, unsigned int selection_size)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	tpos_linked_list_clear(&data->additional_blocked);
	if(data->path){
		free(data->path);
		data->path = NULL;
	}
	unit->is_going = 1;
	data->going_to.x = tx;
	data->going_to.y = ty;
	data->target_reached = 0;
	data->make_them_evade = 0;
	data->count_of_units_going_to_same_place = selection_size;
	return 0;
}
int click_unit_test(struct unit *unit, int ui)
{
	return -1;
}
int going_to_test(struct unit *unit, struct tpos *t)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	t->x = data->going_to.x;
	t->y = data->going_to.y;
	return 0;
}

struct tpos * get_copy_of_path_test(struct unit *unit, struct tpos *starting_from)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	if(data->path == NULL){
		//fprintf(stderr, "get_copy_of_path_test: unit %i hasn't got a path\n", unit->index);
		return NULL;
	}
	struct tpos *t = data->path;
	//search the starting node in the path
	if(starting_from != NULL){
		//fprintf(stderr, "get_copy_of_path_test: unit %i: searching starting_from=(%i,%i) in path:\n", unit->index, starting_from->x, starting_from->y);
		for(;;){
			if(t->x == starting_from->x && t->y == starting_from->y){
				//fprintf(stderr, "is (%i,%i).\n", t->x, t->y);
				break;
			}
			//fprintf(stderr, "not (%i,%i), ", t->x, t->y);
			if(t->x == data->going_to.x && t->y == data->going_to.y){
				//fprintf(stderr, "[end came.]\n");
				t = NULL;
				break;
			}
			t++;
		}
		//if the starting_from node wasn't found, return NULL
		if(t==NULL){
			//fprintf(stderr, "get_copy_of_path_test: unit %i: the node (%i,%i) isn't on unit's path.\n", unit->index, starting_from->x, starting_from->y);
			return NULL;
		}
		//if the starting_from node was the final node, return NULL
		if(t->x == data->going_to.x && t->y == data->going_to.y){
			//fprintf(stderr, "get_copy_of_path_test: unit %i: the node (%i,%i) is the last node on unit's path.\n", unit->index, starting_from->x, starting_from->y);
			return NULL;
		}
		/*if(t->x == data->going_to.x && t->y == data->going_to.y){
			fprintf(stderr, "get_copy_of_path_test: unit %i: the node (%i,%i) isn't on unit's path ( ", unit->index, t->x, t->y);
			t = data->path;
			while(t->x != data->going_to.x && t->y != data->going_to.y){
				fprintf(stderr, "(%i,%i) ", t->x, t->y);
				t++;
			}
			fprintf(stderr, ")\n");
			return NULL;
		}*/
	}
	struct tpos *t2;
	//count the length of the path
	unsigned int length = 0;
	t2 = t;
	for(;;){
		length++;
		if(t2->x == data->going_to.x && t2->y == data->going_to.y){
			break;
		}
		t2++;
	}
	//allocate memory for the copy and copy the path to it
	struct tpos *copy = (struct tpos*)malloc(sizeof(struct tpos)*length);
	if(copy==NULL){
		fprintf(stderr, "get_copy_of_path_test(): could not allocate memory\n");
		assert(0);
		return NULL;
	}
	struct tpos *c = copy;
	t2 = t;
	//fprintf(stderr, "get_copy_of_path_test: unit %i: returning a path of length %i:\n", unit->index, length);
	unsigned int i=0;
	for(;;){
		if(i==length) assert(0);
		c->x = t2->x;
		c->y = t2->y;
		//fprintf(stderr, "(%i,%i) ", c->x, c->y);
		if(t2->x == data->going_to.x && t2->y == data->going_to.y){
			break;
		}
		t2++;
		c++;
		i++;
	}
	//fprintf(stderr, "\n");
	return copy;
}

struct tpos_linked * get_additional_blocked_test(struct unit *unit)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	return data->additional_blocked;
}

void sync_additional_blocked_test(struct unit *unit, struct tpos_linked *a)
{
	//sel_fprintf(stderr, "sync_additional_blocked_test: unit %i\n", unit->index);

	if(a==NULL){
		//sel_fprintf(stderr, "sync_additional_blocked_test: given NULL list\n");
		return;
	}

	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;

	int c_this = tpos_linked_list_count(data->additional_blocked);
	//int c_other= tpos_linked_list_count(a);
	int c_can_add = TEST_ADD_BLOCKED_MAX_COUNT - c_this;
	//sel_fprintf(stderr, "c_this=%i, c_other=%i, c_can_add=%i\n", c_this, c_other, c_can_add);
	int c_added = 0;
	struct tpos_linked *l = a;
	while(l!=NULL && c_added < c_can_add){
		struct tpos_linked *l2 = data->additional_blocked;
		while(l2!=NULL){
			if(l->x == l2->x && l->y == l2->y) break;
			l2 = l2->next;
		}
		if(l2==NULL){
			//sel_fprintf(stderr, "\tadding (%i,%i) to this unit's list.\n", l->x, l->y);
			tpos_linked_list_add_single_copy(&data->additional_blocked, (struct tpos*)l);
			c_added++;
		}
		l = l->next;
	}
	if(c_added == 0){
		//sel_fprintf(stderr, "\tadded nothing to this unit's list.\n")
		return;
	}
	//Added something, so clear the current path.
	free(data->path);
	data->path = NULL;
}

int get_target_reached_test(struct unit *unit, struct tpos *t)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	if(!data->target_reached) return -1;
	t->x = data->going_to.x;
	t->y = data->going_to.y;
	return 0;
}
int evade_test(struct unit *unit, struct game *game, struct unit *caller, unsigned int recursion_depth)
{
	assert(unit->data!=NULL);
	struct data_test *data = (struct data_test *)unit->data;
	if(unit->is_going) return -1;
	//if(unit->is_moving) return -1;
	struct tpos ct; //current tpos
	ct.x = unit->p.x/MAP_TILE_W;
	ct.y = unit->p.y/MAP_TILE_H;
	struct tpos n[4];
	n[0].x = ct.x-1;
	n[0].y = ct.y;
	n[1].x = ct.x+1;
	n[1].y = ct.y;
	n[2].x = ct.x;
	n[2].y = ct.y-1;
	n[3].x = ct.x;
	n[3].y = ct.y+1;
	unsigned int toadd = rand()%4;
	unsigned int k;
	int p = -1;
	for(k=0; k<2 && p == -1; k++){
		unsigned int i;
		for(i=0; i<4 && p == -1; i++){
			int h = (i+toadd)%4;
			struct tile_info ti;
			if(game_get_tile_info(game, &ti, n[h].x, n[h].y)==-1){
				continue;
			}
			if(!(ti.mapflags & MAPTILE_WALKABLE)) continue;
			if(ti.unit_count==0){
				p = h;
				break;
			}
			unsigned int l;
			unsigned int caller_is_at_target = 0;
			for(l=0; l<ti.unit_count; l++){
				struct unit *u = &game->units[ti.unit_i[l]];
				if(u == caller){
					caller_is_at_target = 1;
					break;
				}
			}
			if(caller_is_at_target){
				continue;
			}
			unsigned int everyone_evades = 1;
			for(l=0; l<ti.unit_count; l++){
				struct unit *u = &game->units[ti.unit_i[l]];
				if(unit_evade(u, game, unit, recursion_depth)==-1){
					everyone_evades = 0;
					break;
				}
			}
			if(everyone_evades){
				p = h;
				break;
			}
		}
	}
	if(p == -1){
		return -1;
	}

	//Got a place to evade, which is n[p]. Now get going.

	/*unit->moving_to.x = n[p].x * MAP_TILE_W;
	unit->moving_to.y = n[p].y * MAP_TILE_H;
	unit->is_moving = 1;
	unit->direction = tjump_direction(&ct, &n[p]);*/

	tpos_linked_list_clear(&data->additional_blocked);
	if(data->path){
		free(data->path);
		data->path = NULL;
	}
	unit->is_going = 1;
	data->going_to.x = n[p].x;
	data->going_to.y = n[p].y;
	data->target_reached = 0;
	data->make_them_evade = 0;

	return 0;
}

//=== other functions related to the unit_specs functions ===

int8_t blockedfunc_test(struct tile_info *info)
{
	//if(info->mapflags & MAPTILE_WALKABLE && info->unit_count==0) return 0;
	if(info->mapflags & MAPTILE_WALKABLE) return 0;
	return -1;
}

//=== unit class ===

int unit_co(struct unit *self, uint16_t index)
{
	//fprintf(stderr, "unit_co:\n");
	//fprintf(stderr, "self->type_id=%i\n", self->type_id);
	self->type_id = -1;
	//fprintf(stderr, "self->type_id=%i\n", self->type_id);
	self->data = NULL;
	self->graphics_changed = 0;
	self->selected = 0;
	self->was_selected = 0;
	self->index = index;
	return 0;
}
void unit_de(struct unit *self)
{
	unit_unload(self);
}
int unit_load(struct unit * self, int type_id)
{
	//fprintf(stderr, "unit_load(self, type_id=%i)\n", type_id);
	if(self->type_id != -1){
		fprintf(stderr, "unit_load: can't load on an unit with type_id!=-1\n");
		return -1;
	}
	assert(self->data == NULL);
	self->type_id = type_id;
	self->selected = 0;
	self->graphics_changed = 1;
	self->p.x = 0;
	self->p.y = 0;
	self->is_moving = 0;
	self->is_going = 0;
	self->direction = DIRECTION_NONE;
	self->evade_being_processed = 0;
	if(unit_specs[self->type_id].load != NULL){
		return (*unit_specs[self->type_id].load)(self);
	}
	fprintf(stderr, "unit_load: warning: no load function in unit_spec\n");
	return -1;
}
void unit_unload(struct unit *self)
{
	if(self->type_id != -1){
		assert(self->type_id >=0 && self->type_id <= TYPE_ID_MAX);
		if(unit_specs[self->type_id].unload != NULL){
			(*unit_specs[self->type_id].unload)(self);
		}
	}
	self->type_id = -1;
	self->data = NULL;
}
int unit_get_draw_info(struct unit *self, struct draw_image_info *images, unsigned int maxcount)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_draw_info == NULL) return -1;
	return (*unit_specs[self->type_id].get_draw_info)(self, images, maxcount);
}
int unit_get_last_draw_rects(struct unit *self, SDL_Rect *rects, unsigned int maxcount)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_last_draw_rects == NULL) return -1;
	return (*unit_specs[self->type_id].get_last_draw_rects)(self, rects, maxcount);
}
int unit_set_tpos(struct unit *self, int x, int y, int slot)
{
	//fprintf(stderr, "unit_set_tpos(%i (type_id=%i), %i, %i, %i)\n", (int)self, self->type_id, x, y, slot);
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].set_tpos == NULL) return -1;
	return (*unit_specs[self->type_id].set_tpos)(self, x, y, slot);
}
/*int unit_get_tpos(struct unit *self, int *x, int *y, int *slot)
{
	//fprintf(stderr, "unit_get_tpos\n");
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_tpos == NULL) return -1;
	return (*unit_specs[self->type_id].get_tpos)(self, x, y, slot);
}*/
int unit_get_blocked_tiles(struct unit *self, struct tpos *ts)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_blocked_tiles == NULL) return -1;
	return (*unit_specs[self->type_id].get_blocked_tiles)(self, ts);
}
int unit_action(struct unit *self, struct game *game)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	
	if(self->is_moving){
		//fprintf(stderr, "unit_action: is moving from (%i,%i) to (%i,%i)\n", self->p.x, self->p.y, self->moving_to.x, self->moving_to.y);
		float dx = (float)(self->moving_to.x - self->p.x);
		float dy = (float)(self->moving_to.y - self->p.y);
		float d = sqrt(dx*dx+dy*dy);
		if(abs(d)>0.001){
			//fprintf(stderr, "dx=%f dy=%f, move_speed=%f\n", dx, dy, self->move_speed);
			float x_part = dx/d;
			float y_part = dy/d;
			//fprintf(stderr, "x_part=%f, y_part=%f\n", x_part, y_part);
			self->move_counter_x += self->move_speed * x_part;
			self->move_counter_y += self->move_speed * y_part;
			//fprintf(stderr, "move_counter_x=%f, move_counter_y=%f\n", self->move_counter_x, self->move_counter_y);
			int move_x = (int)self->move_counter_x;
			int move_y = (int)self->move_counter_y;
			if(dx>0 && move_x > dx) move_x = dx;
			if(dx<0 && move_x < dx) move_x = dx;
			if(dy>0 && move_y > dy) move_y = dy;
			if(dy<0 && move_y < dy) move_y = dy;
			//fprintf(stderr, "move_x=%i, move_y=%i\n", move_x, move_y);
			self->p.x += move_x;
			self->move_counter_x -= (float)move_x;
			self->p.y += move_y;
			self->move_counter_y -= (float)move_y;
			self->graphics_changed = 1;
			//fprintf(stderr, "UNIT %i AT (%i,%i)px\n", self->index, self->p.x, self->p.y);
		}
		if(self->moving_to.x == self->p.x && self->moving_to.y == self->p.y){
			self->is_moving = 0;
			self->move_counter_x = 0.0;
			self->move_counter_y = 0.0;
		}
	}

	if(unit_specs[self->type_id].action == NULL) return -1;

	return (*unit_specs[self->type_id].action)(self, game);
}
int unit_get_select_area(struct unit *self, SDL_Rect *rect)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_select_area == NULL) return -1;
	(*unit_specs[self->type_id].get_select_area)(self, rect);
	return 0;
}
int unit_click_map(struct unit *self, int x, int y, unsigned int selection_size)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].click_map == NULL) return -1;
	return (*unit_specs[self->type_id].click_map)(self, x, y, selection_size);
}
int unit_click_unit(struct unit *self, int ui)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].click_unit == NULL) return -1;
	return (*unit_specs[self->type_id].click_unit)(self, ui);
}
int unit_going_to(struct unit *self, struct tpos *t)
{
	if(self->type_id == -1) return -1;
	if(self->is_going == 0) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].going_to == NULL) return -1;
	return (*unit_specs[self->type_id].going_to)(self, t);
}
struct tpos * unit_get_copy_of_path(struct unit *self, struct tpos *starting_from)
{
	if(self->type_id == -1) return NULL;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_copy_of_path == NULL) return NULL;
	return (*unit_specs[self->type_id].get_copy_of_path)(self, starting_from);
}
struct tpos_linked * unit_get_additional_blocked(struct unit *self)
{
	if(self->type_id == -1) return NULL;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_additional_blocked == NULL) return NULL;
	return (*unit_specs[self->type_id].get_additional_blocked)(self);
}
int unit_sync_additional_blocked(struct unit *self, struct tpos_linked *a)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].sync_additional_blocked == NULL) return -1;
	(*unit_specs[self->type_id].sync_additional_blocked)(self, a);
	return 0;
}
int unit_get_target_reached(struct unit *self, struct tpos *t)
{
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].get_target_reached == NULL) return -1;
	return (*unit_specs[self->type_id].get_target_reached)(self, t);
}
int unit_evade(struct unit *self, struct game *game, struct unit *caller, unsigned int recursion_depth)
{
	if(self->evade_being_processed) return -1;
	recursion_depth++;
	if(recursion_depth > UNIT_EVADE_MAX_RECURSION) return -1;
	self->evade_being_processed = 1;
	if(self->type_id == -1) return -1;
	assert(self->type_id>=0 && self->type_id<=TYPE_ID_MAX);
	if(unit_specs[self->type_id].evade == NULL) return -1;
	int i = (*unit_specs[self->type_id].evade)(self, game, caller, recursion_depth);
	self->evade_being_processed = 0;
	return i;
}

