#ifndef UNIT_H
#define UNIT_H

#include "stdint.h"
#include <SDL/SDL.h>
#include "imglib.h"
#include "map.h" //only for ppos
#include "common.h"

#define UNIT_IMAGE_MAX_COUNT 16
#define UNIT_MAX_DRAW_IMAGE_COUNT 2
#define MAX_UNITS_PER_TILE 4
#define TYPE_ID_MAX 127
#define UNIT_MAX_BLOCKED_TILES 12
#define UNIT_EVADE_MAX_RECURSION 3

struct game;
struct unit;

struct draw_image_info
{
	SDL_Surface *s;
	int px, py;
};

struct unit_spec
{
	char * name;
	char * imagenames[UNIT_IMAGE_MAX_COUNT];
	int (*load)(struct unit *unit);
	void (*unload)(struct unit *unit);
	//returns number of images or -1 on error
	int (*get_draw_info)(struct unit *unit, struct draw_image_info *images, unsigned int maxcount);
	/*
		This is used for erasing the images from the old positions.
		It sets *rects to the surrounding rectangles of the images given
		in the last get_draw_info.
	*/
	int (*get_last_draw_rects)(struct unit *unit, SDL_Rect *rects, unsigned int maxcount);
	/*
		These set/get the unit's location on the map in tiles.
		In case the unit is larger than one tile, the upper left corner is used.
		slot:
			-1 if N/A,
			an integer if more than one unit can be fit in one tile.
	*/
	int (*set_tpos)(struct unit *unit, int x, int y, int slot);
	//if slot is NULL, it is not set.
	//int (*get_tpos)(struct unit *unit, int *x, int *y, int *slot);
	/*
		This is used to get the tile positions on which the unit currently is.
		ts: a pointer to an array of struct tpos of size UNIT_MAX_BLOCKED_TILES
		return value: number of tiles used
	*/
	int (*get_blocked_tiles)(struct unit *unit, struct tpos *ts);
	//this is called once in every frame
	int (*action)(struct unit *unit, struct game *game);
	/*
		Gives an area that can be clicked to select the unit.
		The selection rectangle is drawn to this size.
		rect: the area. units in pixels from map's upper left corner
	*/
	void (*get_select_area)(struct unit *unit, SDL_Rect *rect);
	/*
		This is called when the unit is selected and the user clicks somewhere
		on the map background.
		return value:
			-1: the unit will do nothing (so the unit is deselected)
			0: the unit does something (keep selection)
		x,y: position on map in tiles
	*/
	int (*click_map)(struct unit *unit, int tx, int ty, unsigned int selection_size);
	/*
		This is called when the unit is selected and the user clicks an another
		unit.
		return value:
			-1: the unit will do nothing (so the other unit is selected)
			0: the unit does something (keep selection)
		ui: index number to game's units[] array
	*/
	int (*click_unit)(struct unit *unit, int ui);
	/*
		Gives the position where the unit is trying to go in tiles
		return value:
			-1: the unit isn't going anywhere
			0: the unit is going somewhere (and t is set accordingly)
	*/
	int (*going_to)(struct unit *unit, struct tpos *t);
	/*
		Gives the path to where the unit is going, starting from the given
		point.
		starting_from: a valid tile position on the map, or NULL if the whole
		               path is requested
		return value:
			A pointer to a copy of the path (starting from the given point).
			Caller must free() the memory.
			or
			NULL if 1) the unit isn't going anywhere (is_going==0).
			        2) it doesn't have the path functionality.
			        3) the given position is not on the path.
			        4) the given position is the end node of the path.
	*/
	struct tpos * (*get_copy_of_path)(struct unit *unit, struct tpos *starting_from);
	/*
		Returns the additional blocked positions known by the unit for
		pathfinding, or NULL if it knows no additional blocked positions.
	*/
	struct tpos_linked * (*get_additional_blocked)(struct unit *unit);
	/*
		Synchronizes parameter a to the list of additional blocked positions
		of the unit, if applicable.
		The unit should regenerate it's path if it adds something to it's list.
	*/
	void (*sync_additional_blocked)(struct unit *unit, struct tpos_linked *a);
	/*
		This is used to get the target position that the unit considers to
		have reached at the moment.
		return value:
		    0: have reached a target. target position is copied to t
		    -1: haven't reached a target (or doesn't have a target)
	*/
	int (*get_target_reached)(struct unit *unit, struct tpos *t);
	/*
		This makes the unit try to move one tile to somewhere where it can.
		returns 0 if succesful, -1 otherwise.
		Unit should try to evade only if it's staying at the place.
		unit_evade() prevents recursion loops.
		recursion_depth should be set to 0 by the caller
	*/
	int (*evade)(struct unit *unit, struct game *game, struct unit *caller, unsigned int recursion_depth);

	//The following variables are set in unit_specs_initialize():
	//list of pointers to images in imglib
	SDL_Surface *images[UNIT_IMAGE_MAX_COUNT];
};

//this sets the runtime-set variables (run before unit_specs[] is used)
void unit_specs_initialize(struct imglib *imglib);

struct unit
{
	int8_t type_id; //index in unit_specs[]. -1 = unit doesn't exist
	/*
		unit_specs functions change this to 1 when the graphics would be drawn
		differently (different images or different place).
		When drawn on screen, this is set to 0 by the drawer.
	*/
	uint8_t graphics_changed;
	//1 if selected, 0 if not
	uint8_t selected;
	/*
		Unit position on map in pixels.
		To enable moving the unit smoothly by setting is_moving, moving_to
		and move_speed, get_draw_info should return something relative to these.
	*/
	struct ppos p;
	/*
		1 if the unit should be and is moved towards moving_to at move_speed,
		0 otherwise.
	*/
	uint8_t is_moving;
	/*
		The position where the unit should be and is being moved at move_speed,
		if is_moving==1
	*/
	struct ppos moving_to;
	/*
		The speed in pixels per frame at which the unit is moved if is_moving==1
	*/
	float move_speed;
	/*
		These are used by the move routine in unit_action.
	*/
	float move_counter_x, move_counter_y;
	/*
		This is set by the unit's unit_specs functions.
		It should be 1 always when the unit is attempting to go somewhere.
		It should be 0 when the unit is attempting to stand still.
	*/
	uint8_t is_going;
	/*
		This is set by the unit's unit_specs functions.
		It should represent the direction where the unit is currently going (or
		trying to go.)
		If is_going == 0, this can be anything.
		It should be usually set by the path following routine.
		If the unit isn't able to move, it can be left to the default value.
		value:
		if is_going==1:
			DIRECTION_NONE: the unit doesn't know the direction
			<anything else>: the unit is moving to to this direction
		if is_going==0:
			<anything>: value has no meaning.
	*/
	uint8_t direction;
	/*
		This is set when the unit is set unselected and cleared when the unit
		is drawn, so that the drawing routine knows when to clear the selection
		rectangle.
	*/
	uint8_t was_selected;

	uint8_t evade_being_processed;

	//the following variables are set when game is loaded:

	//index number in units[]. This is set in unit_co.
	uint16_t index;

	//the following variables are set when the unit is loaded:

	//Additional data used by the functions in unit_specs.
	void *data;
};

/*
	This has to be called before doing anything else with the unit structure.
	index: index of the unit in game->units[]
*/
int unit_co(struct unit *self, uint16_t index);
//this has to be called before exiting the program for every unit structure
void unit_de(struct unit *self);

/*
	Wrappers for unit_spec functions:
	Look for more info in struct unit_spec.
	Return value:
		-1 or NULL if type_id==-1 or the actual function pointer is NULL
	otherwise the return value of the actual function, or 0 if void
*/

/*
	this loads the unit with unit_specs[type_id].
	before this, the unit doesn't exist in the game
*/
int unit_load(struct unit *self, int type_id);
//this unloads the unit and makes it non-existent
void unit_unload(struct unit *self);
int unit_get_draw_info(struct unit *self, struct draw_image_info *images, unsigned int maxcount);
int unit_get_last_draw_rects(struct unit *self, SDL_Rect *rects, unsigned int maxcount);
//set/get unit position on map in tiles: returns -1 on error
int unit_set_tpos(struct unit *self, int x, int y, int slot);
//int unit_get_tpos(struct unit *self, int *x, int *y, int *slot);
int unit_get_blocked_tiles(struct unit *self, struct tpos *ts);
int unit_action(struct unit *self, struct game *game);
int unit_get_select_area(struct unit *self, SDL_Rect *rect);
int unit_click_map(struct unit *self, int tx, int ty, unsigned int selection_size);
int unit_click_unit(struct unit *self, int ui);
int unit_going_to(struct unit *self, struct tpos *t);
struct tpos * unit_get_copy_of_path(struct unit *self, struct tpos *starting_from);
struct tpos_linked * unit_get_additional_blocked(struct unit *self);
int unit_sync_additional_blocked(struct unit *self, struct tpos_linked *a);
int unit_get_target_reached(struct unit *self, struct tpos *t);
int unit_evade(struct unit *self, struct game *game, struct unit *caller, unsigned int recursion_depth);

#endif

